/*
 * "fbp.c" - code for "frame buffer pointer" object type
 *
 * written by Peter C. Jones
 * last modified 7-18-1994
 */

/************************************************************
Copyright (c) 1994  Sun Microsystems, Inc.
Copyright (c) 1993  X Consortium

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documenta-
tion files (the "Software"), to deal in the Software with-
out restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to
whom the Software is furnished to do so, subject to the 
following conditions:

The above copyright notice and this permission notice shall
be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE, AND NONINFRINGEMENT.  IN NO EVENT SHALL SUN OR THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHER-
WISE, ARISING FROM, OUT OF, OR IN CONNECTION WITH THE 
SOFTWARE OR THE USE OR DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the names of Sun
Microsystems, Inc. and the X Consortium shall not be used in
advertising or otherwise to promote the sale, use, or other
dealings in this Software without prior written authoriza-
tion from Sun Microsystems, Inc. and/or the X Consortium, as
appropriate.
*************************************************************/

#ident "%W% Xvan 7-18-1994"

#include <stdio.h>

extern "C" {
#include "misc.h"
#include "servermd.h"
}

#include "fbp.h"

/*
 * "FBList::FBList()" - constructor for FBList objects
 */
FBList::FBList() {
  num = width = height = depth = 0;
  yard = NULL;
  table = &yard;
}

/*
 * "FBList::~FBList()" - destructor for FBList objects
 */
FBList::~FBList() {
  if (num > 1)
    xfree(table);
  if (yard)
    xfree(yard);
}

/*
 * "FBList::addRel()" - add another frame buffer to this list, given
 *                      relative position to last one added
 */
void FBList::addRel(int w, int h, int d,
		    unsigned int *a, position p)
{
  if (num >= MAXFBS)
    FatalError("FBList::addRel(): too many frame buffers in list");
  if (!num) {
    fbdesc[0].x = 0;
    fbdesc[0].y = 0;
    width = fbdesc[0].right = fbdesc[0].width = w;
    height = fbdesc[0].bottom = fbdesc[0].height = h;
    depth = d;
    fbdesc[0].addr = a;
    wordwidth = (PixmapBytePad(w, d) >> 2);
    tablesize = 1;
    shift = 31;
    mask = 0xFFFFFFFF;
    table = &fbdesc[0].addr;
    table[1] = fbdesc[0].addr;          // HACK: prevent reading over on
                                        // copies from memory pixmaps
    screenswide = screenshigh = 1;
    num = 1;
  }
  else {
    int delta, i, j, k, line, n;
    unsigned int shifted;
    FBDesc *current = &fbdesc[num];
    FBDesc *last = &fbdesc[num - 1];

    if (w != last->width)
      FatalError("FBList:addRel(): attempt to combine frame buffers "
		 "of different width");
    if (h != last->height)
      FatalError("FBList:addRel(): attempt to combine frame buffers "
		 "of different height");
    if (d != depth)
      FatalError("FBList:addRel(): attempt to combine frame buffers "
		 "of different depth");
    
    current->width = w;
    current->height = h;
    current->addr = a;
    switch (p) {
    case RIGHT:
      current->x = last->right;
      current->y = last->y;
      break;
    case BOTTOM:
      current->x = last->x;
      current->y = last->bottom;
      break;
    case LEFT:
      if ((delta = w - last->x) > 0)
	for (i = 0; i < num; ++ i) {
	  fbdesc[i].x += delta;
	  if ((fbdesc[i].right += delta) > width)
	    width = fbdesc[i].right;
	}
      current->x = last->x - w;
      current->y = last->y;
      break;
    case TOP:
      if ((delta = h - last->y) > 0)
	for (i = 0; i < num; ++ i) {
	  fbdesc[i].y += delta;
	  if ((fbdesc[i].bottom += delta) > height)
	    height = fbdesc[i].bottom;
	}
      current->x = last->x;
      current->y = last->y - h;
      break;
    }
    current->right = current->x + w;
    current->bottom = current->y + h;

    if (current->right > width)
      width = current->right;
    if (current->bottom > height)
      height = current->bottom;
    screenswide = width / w;
    screenshigh = height / h;

    for (shift = 0, shifted = wordwidth;
	 !(shifted & 0x1);
	 ++ shift, shifted >>= 1);
    tablesize = shifted * screenswide * height;
    mask = (1 << shift) - 1;
    if (num > 1)
      xfree(table);
    table = (unsigned int **) xalloc(tablesize * sizeof(unsigned int *));
    if (yard)
      xfree(yard);
    yard = (unsigned int *) xalloc((wordwidth / shifted) << 2);
    for (i = 0, line = 0;
	 i < tablesize;
	 i += shifted * screenswide, ++ line) {
      for (k = 0; k < screenswide; ++ k) {
	for (n = 0;; ++ n) {
	  if ((k == (fbdesc[n].x / fbdesc[n].width)) &&
	      (line >= fbdesc[n].y) &&
	      (line < fbdesc[n].bottom))
	    break;
	  if (n >= num) {
	    n = -1;
	    break;
	  }
	}
	for (j = 0; j < shifted; ++ j) {
	  table[i + shifted * k + j] =
	    (n >= 0?
	     (fbdesc[n].addr +
	      (line - fbdesc[n].y) * wordwidth + j * (wordwidth / shifted)) :
	     yard);
	}
      }
    }
    ++ num;
  }
}

/*
 * "FBPBase::FBPBase()" - default constructor for FBPBase objects
 */
FBPBase::FBPBase() {
  type = uninitialized;
}

/*
 * "FBPBase::FBPBase(FBList *)" - constructor of FBPBase object from an FBList
 */
FBPBase::FBPBase(FBList *fblp)
{
  if (fblp->num < 1)
    FatalError("FBP::FBP(FBList *): attached to empty frame buffer list");
  if (fblp->num > 1)
    type = multiple;
  else
    type = simple;
  fblist = fblp;
  table = (unsigned char **) fblp->table;
  if ((shift = fblp->shift + 2) > 31)
    shift = 31;
  mask = (fblp->mask << 2) | 0x3;
  value = 0;
}

/* 
 * "FBPBase::initfake(unsigned int *)" - initialize FBPxx object 
 *                                       from a raw pointer
 */
void FBPBase::initfake(unsigned int *addr)
{
  static FBList fakefblist;
  static int fakefblist_inited = 0;
  if (!fakefblist_inited) {
    fakefblist.shift = 31;
    fakefblist.mask = 0xFFFFFFFF;
    fakefblist.fbdesc[0].addr = 0;
    fakefblist.table = &fakefblist.fbdesc[0].addr;
  }
  type = fake;
  fblist = &fakefblist;
  table = (unsigned char **) fakefblist.table;
  if ((shift = fakefblist.shift + 2) > 31)
    shift = 31;
  mask = (fakefblist.mask << 2) | 0x3;
  value = (unsigned int) addr;
}

/*
 * "FBP::debug_operator_star()" - function for debugging dereferencing operator
 */
unsigned int& FBP::debug_operator_star()
{
  switch (type) {
  case uninitialized:
    FatalError("FBP::debug_operator_star(): uninitialized FBP dereferenced");
    break;
  case fake:
    break;
  default:
    if (value < 0 || (value >> 2) > (fblist->wordwidth * fblist->screenswide *
				     fblist->height))
      FatalError("FBP::debug_operator_star(): FBP dereferenced out of range");
  }
  return *((unsigned int *)
	   (table[value >> shift] + (value & mask)));
}

/*
 * "FBP::debug_operator_brackets()" - function for debugging indexing operator
 */
unsigned int& FBP::debug_operator_brackets(int x)
{
  unsigned int offset = value + (x << 2);

  switch (type) {
  case uninitialized:
    FatalError("FBP::debug_operator_brackets(): "
	       "uninitialized FBP dereferenced");
    break;
  case fake:
    break;
  default:
    if (offset < 0 || 
	(offset >> 2) > (fblist->wordwidth * fblist->screenswide * 
			 fblist->height))
      FatalError("FBP::debug_operator_brackets(): "
                 "FBP dereferenced out of range");
  }
  return *((unsigned int *)
	   (table[offset >> shift] + (offset & mask)));
}
